@@ -1,23 +1,5 @@
 /*
- * $Id: login.c,v 1.5 1997/06/24 14:18:56 morgan Exp morgan $
- *
- * $Log: login.c,v $
- * Revision 1.5  1997/06/24 14:18:56  morgan
- * update for .55
- *
- * Revision 1.4  1997/02/24 05:59:58  morgan
- * delay only if function available
- * 10 second missuse delay (not 60)
- *
- * Revision 1.3  1997/02/11 15:21:30  morgan
- * update for .54 release
- *
- * Revision 1.2  1997/01/29 03:35:03  morgan
- * update for release
- *
- * Revision 1.1  1996/12/01 00:57:40  morgan
- * Initial revision
- *
+ * $Id: login.c,v 1.10 1998/06/13 07:49:20 saw Exp $
  */
 
 /*
@@ -34,8 +16,8 @@
  */
 
 static const char rcsid[] =
-"$Id: login.c,v 1.5 1997/06/24 14:18:56 morgan Exp morgan $\n"
-" - Login application. <morgan@linux.kernel.org>"
+"$Id: login.c,v 1.10 1998/06/13 07:49:20 saw Exp $\n"
+" - Login application. <saw@msu.ru>"
 ;
 
 #define _BSD_SOURCE
@@ -558,6 +540,7 @@ void main(int argc, const char **argv)
     int retval=LOGIN_FALSE, status;
     pid_t child;
     uid_t uid= (uid_t) -1;
+    const char *place, *err_descr;
     
     /*
      * Parse the arguments to login. There are static variables
@@ -732,7 +715,12 @@ void main(int argc, const char **argv)
 	serious_abort("shell failed to execute");
     }
     
-    utmp_open_session(pamh, child);
+    retval = utmp_open_session(pamh, child, &place, &err_descr);
+    if (retval < 0) {
+        exit_program(1, 60, PAM_ABORT, "login: %s: %s\n", place, err_descr);
+    }else if (retval > 0) {
+        (void) fprintf(stderr, "login: %s: %s\n", place, err_descr);
+    }
 
     /*
      * Process is parent here... wait for the child to exit
@@ -744,7 +732,12 @@ void main(int argc, const char **argv)
 	D(("shell returned %d", status));
     }
 
-    utmp_close_session(pamh, child);
+    retval = utmp_close_session(pamh, &place, &err_descr);
+    if (retval < 0) {
+        exit_program(1, 60, PAM_ABORT, "login: %s: %s\n", place, err_descr);
+    }else if (retval > 0) {
+        (void) fprintf(stderr, "login: %s: %s\n", place, err_descr);
+    }
 
     /* Delete the user's credentials. */
     retval = pam_setcred(pamh, PAM_DELETE_CRED);
