@@ -1,5 +1,5 @@
 /* 
- * $Id: su.c,v 1.12 1997/06/24 14:18:06 morgan Exp morgan $
+ * $Id: su.c,v 1.15 1998/06/13 06:39:32 saw Exp $
  *
  * su.c
  *
@@ -11,21 +11,8 @@
  *
  * Rewritten for Linux-PAM by Andrew Morgan <morgan@linux.kernel.org>
  *
- * $Log: su.c,v $
- * Revision 1.12  1997/06/24 14:18:06  morgan
- * update for .55 release
+ * Modified by Andrey V. Savochkin <saw@msu.ru>
  *
- * Revision 1.11  1997/02/24 06:02:03  morgan
- * major overhaul, read diff
- *
- * Revision 1.10  1997/01/29 03:35:39  morgan
- * update for release
- *
- * Revision 1.9  1996/12/15 16:27:27  morgan
- * root can ignore account management
- *
- * Revision 1.8  1996/12/01 00:54:05  morgan
- * extra checks for no user shell
  */
 
 #define ROOT_UID                  0
@@ -79,6 +66,7 @@ static int state;
 #define SU_STATE_PROCESS_UNKILLABLE  6
 #define SU_STATE_TERMINAL_REOWNED    7
 #define SU_STATE_FORKED              8
+#define SU_STATE_UTMP_WRITTEN        9
 
 static void store_terminal_modes(void);
 static int reset_terminal_modes(void);
@@ -105,7 +93,8 @@ static void su_exec_shell(const char *shell, uid_t uid, int is_login
 
 void main(int argc, char *argv[])
 {
-    int retval, is_login, status, final_retval;
+    int retcode, is_login, status;
+    int retval, final_retval; /* PAM_xxx return values */
     const char *command, *user;
     const char *shell=NULL;
     pid_t child;
@@ -129,8 +118,8 @@ void main(int argc, char *argv[])
     /* ------------ parse the argument list ----------- */
 
 #ifdef HAVE_PWDB
-    retval = pwdb_start();
-    if (retval != PWDB_SUCCESS)
+    retcode = pwdb_start();
+    if (retcode != PWDB_SUCCESS)
 	exit_now(1, "su: failed\n");
 #endif
 
@@ -151,9 +140,9 @@ void main(int argc, char *argv[])
      * state are reflected in the change of "state".
      * Random exits are strictly prohibited :-)  (SAW)
      */
-    status = 1;
-    err_descr = NULL;
-    state = SU_STATE_PAM_INITIALIZED;
+    status = 1;                       /* fake exit status of a child */
+    err_descr = NULL;                 /* errors hasn't happened */
+    state = SU_STATE_PAM_INITIALIZED; /* state -- initial */
 
     do {                              /* abuse loop to avoid using goto... */
 
@@ -173,7 +162,7 @@ void main(int argc, char *argv[])
 	if (retval != PAM_SUCCESS) {
 	    if (getuid() == 0) {
 		(void) fprintf(stderr, "Account management:- %s\n(Ignored)\n"
-                               , pam_strerror(pamh,retval));
+                               , pam_strerror(pamh, retval));
 	    } else {
 		break;
 	    }
@@ -183,7 +172,7 @@ void main(int argc, char *argv[])
 	/* open the su-session */
 
 	place = "pam_open_session";
-        retval = pam_open_session(pamh,0);      /* Must take care to close */
+        retval = pam_open_session(pamh, 0);     /* Must take care to close */
 	if (retval != PAM_SUCCESS)
             break;
 	state = SU_STATE_SESSION_OPENED;
@@ -224,12 +213,12 @@ void main(int argc, char *argv[])
 	    break;
 	state = SU_STATE_PROCESS_UNKILLABLE;
 
-        retval = change_terminal_owner(uid, is_login, command != NULL
+        retcode = change_terminal_owner(uid, is_login, command != NULL
                 , &place, &err_descr);
-	if (retval > 0) {
+	if (retcode > 0) {
 	    (void) fprintf(stderr, "su: %s: %s\n", place, err_descr);
-	    err_descr = NULL;
-	}else if (retval < 0)
+	    err_descr = NULL; /* forget about the problem */
+	}else if (retcode < 0)
 	    break;
 	state = SU_STATE_TERMINAL_REOWNED;
 
@@ -276,7 +265,16 @@ void main(int argc, char *argv[])
             if (is_login) {
 		char buf_c = '\0';
 
-		utmp_open_session(pamh, child);  /* new logname of terminal */
+                /* [uw]tmp writing */
+		retcode = utmp_open_session(pamh, child, &place, &err_descr);
+                if (retcode > 0) {
+                    (void) fprintf(stderr, "su: %s: %s\n", place, err_descr);
+                    err_descr = NULL; /* forget about the problem */
+                }else if (retcode < 0) {
+                    kill(child, SIGKILL);
+                    break;
+                }
+                state = SU_STATE_UTMP_WRITTEN;
 
 		close(fds[0]);
 		write(fds[1], &buf_c, 1);
@@ -291,21 +289,25 @@ void main(int argc, char *argv[])
 	    D(("shell returned %d", status));
 	}
 
-	if (is_login) {
-	    utmp_close_session(pamh, child);   /* retrieve original logname */
-	}
-
     }while (0);                       /* abuse loop to avoid using goto... */
 
-    if (retval != PAM_SUCCESS) {
-	(void) fprintf(stderr, "su: %s\n", pam_strerror(pamh,retval));
+    if (retval != PAM_SUCCESS) {      /* PAM has failed */
+	(void) fprintf(stderr, "su: %s\n", pam_strerror(pamh, retval));
 	final_retval = PAM_ABORT;
-    }else if (err_descr != NULL) {
+    }else if (err_descr != NULL) {    /* system error has happened */
 	(void) fprintf(stderr, "su: %s: %s\n", place, err_descr);
 	final_retval = PAM_ABORT;
     }else
 	final_retval = PAM_SUCCESS;
 
+    /* do [uw]tmp cleanup */
+    if (state >= SU_STATE_UTMP_WRITTEN) {
+        /* could happen only if `is_login' */
+        retcode = utmp_close_session(pamh, &place, &err_descr);
+        if (retcode)
+            (void) fprintf(stderr, "su: %s: %s\n", place, err_descr);
+    }
+
     /* return terminal to local control */
     if (state >= SU_STATE_TERMINAL_REOWNED)
 	restore_terminal_owner();
@@ -398,7 +400,7 @@ static int reset_terminal_modes()
 
 /* ------ unexpected signals ------------------ */
 
-static struct sigaction old_int_act, old_quit_act, old_tstp_act;
+static struct sigaction old_int_act, old_quit_act, old_tstp_act, old_pipe_act;
 
 static void disable_terminal_signals()
 {
@@ -423,6 +425,11 @@ static void disable_terminal_signals()
      * shutdown - either way, nothing appears in syslogs.
      */
     sigaction(SIGTSTP, &act, &old_tstp_act);
+    /*
+     * Ignore SIGPIPE. The parent `su' process may print something
+     * on stderr. Killing of the process would be undesired.
+     */
+    sigaction(SIGPIPE, &act, &old_pipe_act);
 }
 
 static void enable_terminal_signals()
@@ -430,6 +437,7 @@ static void enable_terminal_signals()
     sigaction(SIGINT, &old_int_act, NULL);
     sigaction(SIGQUIT, &old_quit_act, NULL);
     sigaction(SIGTSTP, &old_tstp_act, NULL);
+    sigaction(SIGPIPE, &old_pipe_act, NULL);
 }
 
 /* ------ terminal ownership ------------------ */
